@@ -1,74 +1,19 @@
-# -*- coding: utf-8 -*-
-
-from copy import deepcopy
-import heapq
 import itertools
 import math
-from collections import Iterable
+from collections.abc import Iterable
+from copy import deepcopy
 
 import numpy as np
-import sortedcontainers
 import sortedcollections
+import sortedcontainers
 
-from adaptive.learner.base_learner import BaseLearner
+from adaptive.learner.base_learner import BaseLearner, uses_nth_neighbors
 from adaptive.learner.learnerND import volume
 from adaptive.learner.triangulation import simplex_volume_in_embedding
 from adaptive.notebook_integration import ensure_holoviews
 from adaptive.utils import cache_latest
 
 
-def uses_nth_neighbors(n):
-    """Decorator to specify how many neighboring intervals the loss function uses.
-
-    Wraps loss functions to indicate that they expect intervals together
-    with ``n`` nearest neighbors
-
-    The loss function will then receive the data of the N nearest neighbors
-    (``nth_neighbors``) aling with the data of the interval itself in a dict.
-    The `~adaptive.Learner1D` will also make sure that the loss is updated
-    whenever one of the ``nth_neighbors`` changes.
-
-    Examples
-    --------
-
-    The next function is a part of the `curvature_loss_function` function.
-
-    >>> @uses_nth_neighbors(1)
-    ... def triangle_loss(xs, ys):
-    ...    xs = [x for x in xs if x is not None]
-    ...    ys = [y for y in ys if y is not None]
-    ...
-    ...    if len(xs) == 2: # we do not have enough points for a triangle
-    ...        return xs[1] - xs[0]
-    ...
-    ...    N = len(xs) - 2 # number of constructed triangles
-    ...    if isinstance(ys[0], Iterable):
-    ...        pts = [(x, *y) for x, y in zip(xs, ys)]
-    ...        vol = simplex_volume_in_embedding
-    ...    else:
-    ...        pts = [(x, y) for x, y in zip(xs, ys)]
-    ...        vol = volume
-    ...    return sum(vol(pts[i:i+3]) for i in range(N)) / N
-
-    Or you may define a loss that favours the (local) minima of a function,
-    assuming that you know your function will have a single float as output.
-
-    >>> @uses_nth_neighbors(1)
-    ... def local_minima_resolving_loss(xs, ys):
-    ...     dx = xs[2] - xs[1] # the width of the interval of interest
-    ...
-    ...     if not ((ys[0] is not None and ys[0] > ys[1])
-    ...         or (ys[3] is not None and ys[3] > ys[2])):
-    ...         return loss * 100
-    ...
-    ...     return loss
-    """
-    def _wrapped(loss_per_interval):
-        loss_per_interval.nth_neighbors = n
-        return loss_per_interval
-    return _wrapped
-
-
 @uses_nth_neighbors(0)
 def uniform_loss(xs, ys):
     """Loss function that samples the domain uniformly.
@@ -99,7 +44,7 @@ def default_loss(xs, ys):
     """
     dx = xs[1] - xs[0]
     if isinstance(ys[0], Iterable):
-        dy = [abs(a-b) for a, b in zip(*ys)]
+        dy = [abs(a - b) for a, b in zip(*ys)]
         return np.hypot(dx, dy).max()
     else:
         dy = ys[1] - ys[0]
@@ -111,17 +56,17 @@ def triangle_loss(xs, ys):
     xs = [x for x in xs if x is not None]
     ys = [y for y in ys if y is not None]
 
-    if len(xs) == 2: # we do not have enough points for a triangle
+    if len(xs) == 2:  # we do not have enough points for a triangle
         return xs[1] - xs[0]
 
-    N = len(xs) - 2 # number of constructed triangles
+    N = len(xs) - 2  # number of constructed triangles
     if isinstance(ys[0], Iterable):
         pts = [(x, *y) for x, y in zip(xs, ys)]
         vol = simplex_volume_in_embedding
     else:
         pts = [(x, y) for x, y in zip(xs, ys)]
         vol = volume
-    return sum(vol(pts[i:i+3]) for i in range(N)) / N
+    return sum(vol(pts[i : i + 3]) for i in range(N)) / N
 
 
 def curvature_loss_function(area_factor=1, euclid_factor=0.02, horizontal_factor=0.02):
@@ -129,14 +74,17 @@ def curvature_loss_function(area_factor=1, euclid_factor=0.02, horizontal_factor
     @uses_nth_neighbors(1)
     def curvature_loss(xs, ys):
         xs_middle = xs[1:3]
-        ys_middle = xs[1:3]
+        ys_middle = ys[1:3]
 
         triangle_loss_ = triangle_loss(xs, ys)
         default_loss_ = default_loss(xs_middle, ys_middle)
         dx = xs_middle[1] - xs_middle[0]
-        return (area_factor * (triangle_loss_**0.5)
-                + euclid_factor * default_loss_
-                + horizontal_factor * dx)
+        return (
+            area_factor * (triangle_loss_ ** 0.5)
+            + euclid_factor * default_loss_
+            + horizontal_factor * dx
+        )
+
     return curvature_loss
 
 
@@ -158,8 +106,7 @@ def _get_neighbors_from_list(xs):
     xs_right = np.roll(xs, -1).tolist()
     xs_left[0] = None
     xs_right[-1] = None
-    neighbors = {x: [x_L, x_R] for x, x_L, x_R
-                 in zip(xs, xs_left, xs_right)}
+    neighbors = {x: [x_L, x_R] for x, x_L, x_R in zip(xs, xs_left, xs_right)}
     return sortedcontainers.SortedDict(neighbors)
 
 
@@ -219,14 +166,13 @@ class Learner1D(BaseLearner):
     def __init__(self, function, bounds, loss_per_interval=None):
         self.function = function
 
-        if hasattr(loss_per_interval, 'nth_neighbors'):
+        if hasattr(loss_per_interval, "nth_neighbors"):
             self.nth_neighbors = loss_per_interval.nth_neighbors
         else:
             self.nth_neighbors = 0
 
         self.loss_per_interval = loss_per_interval or default_loss
 
-
         # When the scale changes by a factor 2, the losses are
         # recomputed. This is tunable such that we can test
         # the learners behavior in the tests.
@@ -376,16 +322,14 @@ class Learner1D(BaseLearner):
             self.losses_combined[x, b] = (b - x) * loss / dx
 
         # (no real point left of x) or (no real point right of a)
-        left_loss_is_unknown = ((x_left is None) or
-                                (not real and x_right is None))
+        left_loss_is_unknown = (x_left is None) or (not real and x_right is None)
         if (a is not None) and left_loss_is_unknown:
-            self.losses_combined[a, x] = float('inf')
+            self.losses_combined[a, x] = float("inf")
 
         # (no real point right of x) or (no real point left of b)
-        right_loss_is_unknown = ((x_right is None) or
-                                 (not real and x_left is None))
+        right_loss_is_unknown = (x_right is None) or (not real and x_left is None)
         if (b is not None) and right_loss_is_unknown:
-            self.losses_combined[x, b] = float('inf')
+            self.losses_combined[x, b] = float("inf")
 
     @staticmethod
     def _find_neighbors(x, neighbors):
@@ -436,8 +380,10 @@ class Learner1D(BaseLearner):
             # The point is already evaluated before
             return
         if y is None:
-            raise TypeError("Y-value may not be None, use learner.tell_pending(x)"
-                "to indicate that this value is currently being calculated")
+            raise TypeError(
+                "Y-value may not be None, use learner.tell_pending(x)"
+                "to indicate that this value is currently being calculated"
+            )
 
         # either it is a float/int, if not, try casting to a np.array
         if not isinstance(y, (float, int)):
@@ -505,7 +451,8 @@ class Learner1D(BaseLearner):
         # Find the intervals for which the losses should be calculated.
         intervals, intervals_combined = [
             [(x_m, x_r) for x_m, (x_l, x_r) in neighbors.items()][:-1]
-            for neighbors in (self.neighbors, self.neighbors_combined)]
+            for neighbors in (self.neighbors, self.neighbors_combined)
+        ]
 
         # The the losses for the "real" intervals.
         self.losses = loss_manager(self._scale[0])
@@ -564,8 +511,11 @@ class Learner1D(BaseLearner):
             return [], []
 
         # If the bounds have not been chosen yet, we choose them first.
-        missing_bounds = [b for b in self.bounds if b not in self.data
-                          and b not in self.pending_points]
+        missing_bounds = [
+            b
+            for b in self.bounds
+            if b not in self.data and b not in self.pending_points
+        ]
 
         if len(missing_bounds) >= n:
             return missing_bounds[:n], [np.inf] * n
@@ -579,8 +529,10 @@ class Learner1D(BaseLearner):
         if len(missing_bounds) > 0:
             # There is at least one point in between the bounds.
             all_points = list(self.data.keys()) + list(self.pending_points)
-            intervals = [(self.bounds[0], min(all_points)),
-                         (max(all_points), self.bounds[1])]
+            intervals = [
+                (self.bounds[0], min(all_points)),
+                (max(all_points), self.bounds[1]),
+            ]
             for interval, bound in zip(intervals, self.bounds):
                 if bound in missing_bounds:
                     quals[(*interval, 1)] = np.inf
@@ -591,25 +543,30 @@ class Learner1D(BaseLearner):
         i, i_max = 0, len(self.losses_combined)
         for _ in range(points_to_go):
             qual, loss_qual = quals.peekitem(0) if quals else (None, 0)
-            ival, loss_ival = self.losses_combined.peekitem(i) if i < i_max else (None, 0)
+            ival, loss_ival = (
+                self.losses_combined.peekitem(i) if i < i_max else (None, 0)
+            )
 
-            if (qual is None
-                or (ival is not None
-                    and self._loss(self.losses_combined, ival)
-                        >= self._loss(quals, qual))):
+            if qual is None or (
+                ival is not None
+                and self._loss(self.losses_combined, ival) >= self._loss(quals, qual)
+            ):
                 i += 1
                 quals[(*ival, 2)] = loss_ival / 2
             else:
                 quals.pop(qual, None)
                 *xs, n = qual
-                quals[(*xs, n+1)] = loss_qual * n / (n+1)
+                quals[(*xs, n + 1)] = loss_qual * n / (n + 1)
 
-        points = list(itertools.chain.from_iterable(
-            linspace(*ival, n) for (*ival, n) in quals))
+        points = list(
+            itertools.chain.from_iterable(linspace(*ival, n) for (*ival, n) in quals)
+        )
 
-        loss_improvements = list(itertools.chain.from_iterable(
-            itertools.repeat(quals[x0, x1, n], n - 1)
-            for (x0, x1, n)  in quals))
+        loss_improvements = list(
+            itertools.chain.from_iterable(
+                itertools.repeat(quals[x0, x1, n], n - 1) for (x0, x1, n) in quals
+            )
+        )
 
         # add the missing bounds
         points = missing_bounds + points
@@ -621,24 +578,35 @@ class Learner1D(BaseLearner):
         loss = mapping[ival]
         return finite_loss(ival, loss, self._scale[0])
 
-    def plot(self):
+    def plot(self, *, scatter_or_line="scatter"):
         """Returns a plot of the evaluated data.
 
+        Parameters
+        ----------
+        scatter_or_line : str, default: "scatter"
+            Plot as a scatter plot ("scatter") or a line plot ("line").
+
         Returns
         -------
-        plot : `holoviews.element.Scatter` (if vdim=1)\
-               else `holoviews.element.Path`
+        plot : `holoviews.Overlay`
             Plot of the evaluated data.
         """
+        if scatter_or_line not in ("scatter", "line"):
+            raise ValueError("scatter_or_line must be 'scatter' or 'line'")
         hv = ensure_holoviews()
-        if not self.data:
-            p = hv.Scatter([]) * hv.Path([])
-        elif not self.vdim > 1:
-            p = hv.Scatter(self.data) * hv.Path([])
-        else:
-            xs, ys = zip(*sorted(self.data.items()))
-            p = hv.Path((xs, ys)) * hv.Scatter([])
 
+        xs, ys = zip(*sorted(self.data.items())) if self.data else ([], [])
+        if scatter_or_line == "scatter":
+            if self.vdim == 1:
+                plots = [hv.Scatter((xs, ys))]
+            else:
+                plots = [hv.Scatter((xs, _ys)) for _ys in np.transpose(ys)]
+        else:
+            plots = [hv.Path((xs, ys))]
+
+        # Put all plots in an Overlay because a DynamicMap can't handle changing
+        # datatypes, e.g. when `vdim` isn't yet known and the live_plot is running.
+        p = hv.Overlay(plots)
         # Plot with 5% empty margins such that the boundary points are visible
         margin = 0.05 * (self.bounds[1] - self.bounds[0])
         plot_bounds = (self.bounds[0] - margin, self.bounds[1] + margin)
@@ -654,27 +622,33 @@ class Learner1D(BaseLearner):
         return self.data
 
     def _set_data(self, data):
-        self.tell_many(*zip(*data.items()))
+        if data:
+            self.tell_many(*zip(*data.items()))
 
+    def __getstate__(self):
+        return (
+            self.function,
+            self.bounds,
+            self.loss_per_interval,
+            dict(self.losses),  # SortedDict cannot be pickled
+            dict(self.losses_combined),  # ItemSortedDict cannot be pickled
+            self._get_data(),
+        )
 
-def _fix_deepcopy(sorted_dict, x_scale):
-    # XXX: until https://github.com/grantjenks/sortedcollections/issues/5 is fixed
-    import types
-    def __deepcopy__(self, memo):
-        items = deepcopy(list(self.items()))
-        lm = loss_manager(self.x_scale)
-        lm.update(items)
-        return lm
-    sorted_dict.x_scale = x_scale
-    sorted_dict.__deepcopy__ = types.MethodType(__deepcopy__, sorted_dict)
+    def __setstate__(self, state):
+        function, bounds, loss_per_interval, losses, losses_combined, data = state
+        self.__init__(function, bounds, loss_per_interval)
+        self._set_data(data)
+        self.losses.update(losses)
+        self.losses_combined.update(losses_combined)
 
 
 def loss_manager(x_scale):
     def sort_key(ival, loss):
         loss, ival = finite_loss(ival, loss, x_scale)
         return -loss, ival
+
     sorted_dict = sortedcollections.ItemSortedDict(sort_key)
-    _fix_deepcopy(sorted_dict, x_scale)
     return sorted_dict
 
 
@@ -683,7 +657,7 @@ def finite_loss(ival, loss, x_scale):
     sort intervals that have infinite loss."""
     # If the loss is infinite we return the
     # distance between the two points.
-    if math.isinf(loss):
+    if math.isinf(loss) or math.isnan(loss):
         loss = (ival[1] - ival[0]) / x_scale
         if len(ival) == 3:
             # Used when constructing quals. Last item is
