@@ -1,20 +1,20 @@
-# -*- coding: utf-8 -*-
-
+import itertools
+import warnings
 from collections import OrderedDict
 from copy import copy
-import itertools
 from math import sqrt
 
 import numpy as np
 from scipy import interpolate
 
 from adaptive.learner.base_learner import BaseLearner
+from adaptive.learner.triangulation import simplex_volume_in_embedding
 from adaptive.notebook_integration import ensure_holoviews
 from adaptive.utils import cache_latest
 
-
 # Learner2D and helper functions.
 
+
 def deviations(ip):
     """Returns the deviation of the linear estimate.
 
@@ -26,12 +26,13 @@ def deviations(ip):
 
     Returns
     -------
-    numpy array
+    deviations : list
         The deviation per triangle.
     """
     values = ip.values / (ip.values.ptp(axis=0).max() or 1)
     gradients = interpolate.interpnd.estimate_gradients_2d_global(
-        ip.tri, values, tol=1e-6)
+        ip.tri, values, tol=1e-6
+    )
 
     p = ip.tri.points[ip.tri.vertices]
     vs = values[ip.tri.vertices]
@@ -40,8 +41,9 @@ def deviations(ip):
     def deviation(p, v, g):
         dev = 0
         for j in range(3):
-            vest = v[:, j, None] + ((p[:, :, :] - p[:, j, None, :]) *
-                                    g[:, j, None, :]).sum(axis=-1)
+            vest = v[:, j, None] + (
+                (p[:, :, :] - p[:, j, None, :]) * g[:, j, None, :]
+            ).sum(axis=-1)
             dev += abs(vest - v).max(axis=1)
         return dev
 
@@ -62,7 +64,7 @@ def areas(ip):
 
     Returns
     -------
-    numpy array
+    areas : numpy.ndarray
         The area per triangle in ``ip.tri``.
     """
     p = ip.tri.points[ip.tri.vertices]
@@ -76,6 +78,15 @@ def uniform_loss(ip):
 
     Works with `~adaptive.Learner2D` only.
 
+    Parameters
+    ----------
+    ip : `scipy.interpolate.LinearNDInterpolator` instance
+
+    Returns
+    -------
+    losses : numpy.ndarray
+        Loss per triangle in ``ip.tri``.
+
     Examples
     --------
     >>> from adaptive.learner.learner2D import uniform_loss
@@ -83,9 +94,11 @@ def uniform_loss(ip):
     ...     x, y = xy
     ...     return x**2 + y**2
     >>>
-    >>> learner = adaptive.Learner2D(f,
-    ...                              bounds=[(-1, -1), (1, 1)],
-    ...                              loss_per_triangle=uniform_loss)
+    >>> learner = adaptive.Learner2D(
+    ...     f,
+    ...     bounds=[(-1, -1), (1, 1)],
+    ...     loss_per_triangle=uniform_loss,
+    ... )
     >>>
     """
     return np.sqrt(areas(ip))
@@ -100,6 +113,10 @@ def resolution_loss_function(min_distance=0, max_distance=1):
     The arguments `min_distance` and `max_distance` should be in between 0 and 1
     because the total area is normalized to 1.
 
+    Returns
+    -------
+    loss_function : callable
+
     Examples
     --------
     >>> def f(xy):
@@ -107,33 +124,41 @@ def resolution_loss_function(min_distance=0, max_distance=1):
     ...     return x**2 + y**2
     >>>
     >>> loss = resolution_loss_function(min_distance=0.01, max_distance=1)
-    >>> learner = adaptive.Learner2D(f,
-    ...                              bounds=[(-1, -1), (1, 1)],
-    ...                              loss_per_triangle=loss)
-    >>>
+    >>> learner = adaptive.Learner2D(f, bounds=[(-1, -1), (1, 1)], loss_per_triangle=loss)
     """
+
     def resolution_loss(ip):
         loss = default_loss(ip)
 
         A = areas(ip)
         # Setting areas with a small area to zero such that they won't be chosen again
-        loss[A < min_distance**2] = 0
+        loss[A < min_distance ** 2] = 0
 
         # Setting triangles that have a size larger than max_distance to infinite loss
         # such that these triangles will be picked
-        loss[A > max_distance**2] = np.inf
+        loss[A > max_distance ** 2] = np.inf
 
         return loss
+
     return resolution_loss
 
 
 def minimize_triangle_surface_loss(ip):
-    """Loss function that is similar to the default loss function in the
+    """Loss function that is similar to the distance loss function in the
     `~adaptive.Learner1D`. The loss is the area spanned by the 3D
     vectors of the vertices.
 
     Works with `~adaptive.Learner2D` only.
 
+    Parameters
+    ----------
+    ip : `scipy.interpolate.LinearNDInterpolator` instance
+
+    Returns
+    -------
+    losses : numpy.ndarray
+        Loss per triangle in ``ip.tri``.
+
     Examples
     --------
     >>> from adaptive.learner.learner2D import minimize_triangle_surface_loss
@@ -165,6 +190,19 @@ def minimize_triangle_surface_loss(ip):
 
 
 def default_loss(ip):
+    """Loss function that combines `deviations` and `areas` of the triangles.
+
+    Works with `~adaptive.Learner2D` only.
+
+    Parameters
+    ----------
+    ip : `scipy.interpolate.LinearNDInterpolator` instance
+
+    Returns
+    -------
+    losses : numpy.ndarray
+        Loss per triangle in ``ip.tri``.
+    """
     dev = np.sum(deviations(ip), axis=0)
     A = areas(ip)
     losses = dev * np.sqrt(A) + 0.3 * A
@@ -182,15 +220,15 @@ def choose_point_in_triangle(triangle, max_badness):
 
     Parameters
     ----------
-    triangle : numpy array
-        The coordinates of a triangle with shape (3, 2)
+    triangle : numpy.ndarray
+        The coordinates of a triangle with shape (3, 2).
     max_badness : int
         The badness at which the point is either chosen on a edge or
         in the middle.
 
     Returns
     -------
-    point : numpy array
+    point : numpy.ndarray
         The x and y coordinate of the suggested new point.
     """
     a, b, c = triangle
@@ -200,7 +238,7 @@ def choose_point_in_triangle(triangle, max_badness):
     i = edge_lengths.argmax()
 
     # We multiply by sqrt(3) / 4 such that a equilateral triangle has badness=1
-    badness = (edge_lengths[i]**2 / area) * (sqrt(3) / 4)
+    badness = (edge_lengths[i] ** 2 / area) * (sqrt(3) / 4)
     if badness > max_badness:
         point = (triangle_roll[i] + triangle[i]) / 2
     else:
@@ -208,6 +246,45 @@ def choose_point_in_triangle(triangle, max_badness):
     return point
 
 
+def triangle_loss(ip):
+    r"""Computes the average of the volumes of the simplex combined with each
+    neighbouring point.
+
+    Parameters
+    ----------
+    ip : `scipy.interpolate.LinearNDInterpolator` instance
+
+    Returns
+    -------
+    triangle_loss : list
+        The mean volume per triangle.
+
+    Notes
+    -----
+    This loss function is *extremely* slow. It is here because it gives the
+    same result as the `adaptive.LearnerND`\s
+    `~adaptive.learner.learnerND.triangle_loss`.
+    """
+    tri = ip.tri
+
+    def get_neighbors(i, ip):
+        n = np.array([tri.simplices[n] for n in tri.neighbors[i] if n != -1])
+        # remove the vertices that are in the simplex
+        c = np.setdiff1d(n.reshape(-1), tri.simplices[i])
+        return np.concatenate((tri.points[c], ip.values[c]), axis=-1)
+
+    simplices = np.concatenate(
+        [tri.points[tri.simplices], ip.values[tri.simplices]], axis=-1
+    )
+    neighbors = [get_neighbors(i, ip) for i in range(len(tri.simplices))]
+
+    return [
+        sum(simplex_volume_in_embedding(np.vstack([simplex, n])) for n in neighbors[i])
+        / len(neighbors[i])
+        for i, simplex in enumerate(simplices)
+    ]
+
+
 class Learner2D(BaseLearner):
     """Learns and predicts a function 'f: ℝ^2 → ℝ^N'.
 
@@ -226,7 +303,6 @@ class Learner2D(BaseLearner):
         triangle area, to determine the loss. See the notes
         for more details.
 
-
     Attributes
     ----------
     data : dict
@@ -244,12 +320,6 @@ class Learner2D(BaseLearner):
         triangles will be stretched along ``x``, otherwise
         along ``y``.
 
-    Methods
-    -------
-    data_combined : dict
-        Sampled points and values so far including
-        the unknown interpolated points in `pending_points`.
-
     Notes
     -----
     Adapted from an initial implementation by Pauli Virtanen.
@@ -337,8 +407,41 @@ class Learner2D(BaseLearner):
 
     @property
     def bounds_are_done(self):
-        return not any((p in self.pending_points or p in self._stack)
-                       for p in self._bounds_points)
+        return not any(
+            (p in self.pending_points or p in self._stack) for p in self._bounds_points
+        )
+
+    def interpolated_on_grid(self, n=None):
+        """Get the interpolated data on a grid.
+
+        Parameters
+        ----------
+        n : int, optional
+            Number of points in x and y. If None (default) this number is
+            evaluated by looking at the size of the smallest triangle.
+
+        Returns
+        -------
+        xs : 1D numpy.ndarray
+        ys : 1D numpy.ndarray
+        interpolated_on_grid : 2D numpy.ndarray
+        """
+        ip = self.interpolator(scaled=True)
+        if n is None:
+            # Calculate how many grid points are needed.
+            # factor from A=√3/4 * a² (equilateral triangle)
+            n = int(0.658 / sqrt(areas(ip).min()))
+            n = max(n, 10)
+
+        # The bounds of the linspace should be (-0.5, 0.5) but because of
+        # numerical precision problems it could (for example) be
+        # (-0.5000000000000001, 0.49999999999999983), then any point at exact
+        # boundary would be outside of the domain. See #181.
+        eps = 1e-13
+        xs = ys = np.linspace(-0.5 + eps, 0.5 - eps, n)
+        zs = ip(xs[:, None], ys[None, :] * self.aspect_ratio).squeeze()
+        xs, ys = self._unscale(np.vstack([xs, ys]).T).T
+        return xs, ys, zs
 
     def _data_in_bounds(self):
         if self.data:
@@ -353,7 +456,8 @@ class Learner2D(BaseLearner):
         if self.pending_points:
             points = list(self.pending_points)
             if self.bounds_are_done:
-                values = self.ip()(self._scale(points))
+                ip = self.interpolator(scaled=True)
+                values = ip(self._scale(points))
             else:
                 # Without the bounds the interpolation cannot be done properly,
                 # so we just set everything to zero.
@@ -370,31 +474,54 @@ class Learner2D(BaseLearner):
         values_combined = np.vstack([values, values_interp])
         return points_combined, values_combined
 
-    def data_combined(self):
-        """Like `data`, however this includes the points in
-        `pending_points` for which the values are interpolated."""
-        # Interpolate the unfinished points
-        points, values = self._data_combined()
-        return {tuple(k): v for k, v in zip(points, values)}
-
     def ip(self):
-        """A `scipy.interpolate.LinearNDInterpolator` instance
-        containing the learner's data."""
-        if self._ip is None:
-            points, values = self._data_in_bounds()
-            points = self._scale(points)
-            self._ip = interpolate.LinearNDInterpolator(points, values)
-        return self._ip
+        """Deprecated, use `self.interpolator(scaled=True)`"""
+        warnings.warn(
+            "`learner.ip()` is deprecated, use `learner.interpolator(scaled=True)`."
+            " This will be removed in v1.0.",
+            DeprecationWarning,
+        )
+        return self.interpolator(scaled=True)
 
-    def ip_combined(self):
+    def interpolator(self, *, scaled=False):
+        """A `scipy.interpolate.LinearNDInterpolator` instance
+        containing the learner's data.
+
+        Parameters
+        ----------
+        scaled : bool
+            Use True if all points are inside the
+            unit-square [(-0.5, 0.5), (-0.5, 0.5)] or False if
+            the data points are inside the ``learner.bounds``.
+
+        Returns
+        -------
+        interpolator : `scipy.interpolate.LinearNDInterpolator`
+
+        Examples
+        --------
+        >>> xs, ys = [np.linspace(*b, num=100) for b in learner.bounds]
+        >>> ip = learner.interpolator()
+        >>> zs = ip(xs[:, None], ys[None, :])
+        """
+        if scaled:
+            if self._ip is None:
+                points, values = self._data_in_bounds()
+                points = self._scale(points)
+                self._ip = interpolate.LinearNDInterpolator(points, values)
+            return self._ip
+        else:
+            points, values = self._data_in_bounds()
+            return interpolate.LinearNDInterpolator(points, values)
+
+    def _interpolator_combined(self):
         """A `scipy.interpolate.LinearNDInterpolator` instance
         containing the learner's data *and* interpolated data of
         the `pending_points`."""
         if self._ip_combined is None:
             points, values = self._data_combined()
             points = self._scale(points)
-            self._ip_combined = interpolate.LinearNDInterpolator(points,
-                                                                 values)
+            self._ip_combined = interpolate.LinearNDInterpolator(points, values)
         return self._ip_combined
 
     def inside_bounds(self, xy):
@@ -424,7 +551,7 @@ class Learner2D(BaseLearner):
             raise ValueError("too few points...")
 
         # Interpolate
-        ip = self.ip_combined()
+        ip = self._interpolator_combined()
 
         losses = self.loss_per_triangle(ip)
 
@@ -437,10 +564,12 @@ class Learner2D(BaseLearner):
             point_new = tuple(self._unscale(point_new))
 
             # np.clip results in numerical precision problems
-            # https://gitlab.kwant-project.org/qt/adaptive/issues/132
-            clip = lambda x, l, u: max(l, min(u, x))
-            point_new = (clip(point_new[0], *self.bounds[0]),
-                         clip(point_new[1], *self.bounds[1]))
+            # https://github.com/python-adaptive/adaptive/issues/7
+            clip = lambda x, l, u: max(l, min(u, x))  # noqa: E731
+            point_new = (
+                clip(point_new[0], *self.bounds[0]),
+                clip(point_new[1], *self.bounds[1]),
+            )
 
             loss_new = losses[jsimplex]
 
@@ -470,7 +599,8 @@ class Learner2D(BaseLearner):
             # than the number of triangles between the points. Therefore
             # it could fill up till a length smaller than `stack_till`.
             new_points, new_loss_improvements = self._fill_stack(
-                stack_till=max(n_left, self.stack_size))
+                stack_till=max(n_left, self.stack_size)
+            )
             for p in new_points[:n_left]:
                 self.tell_pending(p)
             n_left -= len(new_points)
@@ -479,8 +609,7 @@ class Learner2D(BaseLearner):
             loss_improvements += new_loss_improvements
 
         if not tell_pending:
-            self._stack = OrderedDict(zip(points[:self.stack_size],
-                                          loss_improvements))
+            self._stack = OrderedDict(zip(points[: self.stack_size], loss_improvements))
             for point in points[:n]:
                 self.pending_points.discard(point)
 
@@ -490,7 +619,7 @@ class Learner2D(BaseLearner):
     def loss(self, real=True):
         if not self.bounds_are_done:
             return np.inf
-        ip = self.ip() if real else self.ip_combined()
+        ip = self.interpolator(scaled=True) if real else self._interpolator_combined()
         losses = self.loss_per_triangle(ip)
         return losses.max()
 
@@ -501,7 +630,7 @@ class Learner2D(BaseLearner):
                 self._stack[p] = np.inf
 
     def plot(self, n=None, tri_alpha=0):
-        """Plot the Learner2D's current state.
+        r"""Plot the Learner2D's current state.
 
         This plot function interpolates the data on a regular grid.
         The gridspacing is evaluated by checking the size of the smallest
@@ -535,30 +664,26 @@ class Learner2D(BaseLearner):
         lbrt = x[0], y[0], x[1], y[1]
 
         if len(self.data) >= 4:
-            ip = self.ip()
-
-            if n is None:
-                # Calculate how many grid points are needed.
-                # factor from A=√3/4 * a² (equilateral triangle)
-                n = int(0.658 / sqrt(areas(ip).min()))
-                n = max(n, 10)
-
-            x = y = np.linspace(-0.5, 0.5, n)
-            z = ip(x[:, None], y[None, :] * self.aspect_ratio).squeeze()
+            ip = self.interpolator(scaled=True)
+            x, y, z = self.interpolated_on_grid(n)
 
             if self.vdim > 1:
-                ims = {i: hv.Image(np.rot90(z[:, :, i]), bounds=lbrt)
-                       for i in range(z.shape[-1])}
+                ims = {
+                    i: hv.Image(np.rot90(z[:, :, i]), bounds=lbrt)
+                    for i in range(z.shape[-1])
+                }
                 im = hv.HoloMap(ims)
             else:
                 im = hv.Image(np.rot90(z), bounds=lbrt)
 
             if tri_alpha:
                 points = self._unscale(ip.tri.points[ip.tri.vertices])
-                points = np.pad(points[:, [0, 1, 2, 0], :],
-                                pad_width=((0, 0), (0, 1), (0, 0)),
-                                mode='constant',
-                                constant_values=np.nan).reshape(-1, 2)
+                points = np.pad(
+                    points[:, [0, 1, 2, 0], :],
+                    pad_width=((0, 0), (0, 1), (0, 0)),
+                    mode="constant",
+                    constant_values=np.nan,
+                ).reshape(-1, 2)
                 tris = hv.EdgePaths([points])
             else:
                 tris = hv.EdgePaths([])
@@ -566,7 +691,7 @@ class Learner2D(BaseLearner):
             im = hv.Image([], bounds=lbrt)
             tris = hv.EdgePaths([])
 
-        im_opts = dict(cmap='viridis')
+        im_opts = dict(cmap="viridis")
         tri_opts = dict(line_width=0.5, alpha=tri_alpha)
         no_hover = dict(plot=dict(inspection_policy=None, tools=[]))
 
@@ -581,3 +706,18 @@ class Learner2D(BaseLearner):
         for point in copy(self._stack):
             if point in self.data:
                 self._stack.pop(point)
+
+    def __getstate__(self):
+        return (
+            self.function,
+            self.bounds,
+            self.loss_per_triangle,
+            self._stack,
+            self._get_data(),
+        )
+
+    def __setstate__(self, state):
+        function, bounds, loss_per_triangle, _stack, data = state
+        self.__init__(function, bounds, loss_per_triangle)
+        self._set_data(data)
+        self._stack = _stack
