@@ -1,22 +1,30 @@
-# -*- coding: utf-8 -*-
-
-from collections import OrderedDict, Iterable
 import functools
-import heapq
 import itertools
 import random
+from collections import OrderedDict
+from collections.abc import Iterable
 
 import numpy as np
-from scipy import interpolate
 import scipy.spatial
+from scipy import interpolate
 from sortedcontainers import SortedKeyList
 
-from adaptive.learner.base_learner import BaseLearner
-from adaptive.notebook_integration import ensure_holoviews, ensure_plotly
+from adaptive.learner.base_learner import BaseLearner, uses_nth_neighbors
 from adaptive.learner.triangulation import (
-    Triangulation, point_in_simplex, circumsphere,
-    simplex_volume_in_embedding, fast_det)
-from adaptive.utils import restore, cache_latest
+    Triangulation,
+    circumsphere,
+    fast_det,
+    point_in_simplex,
+    simplex_volume_in_embedding,
+)
+from adaptive.notebook_integration import ensure_holoviews, ensure_plotly
+from adaptive.utils import cache_latest, restore
+
+
+def to_list(inp):
+    if isinstance(inp, Iterable):
+        return list(inp)
+    return [inp]
 
 
 def volume(simplex, ys=None):
@@ -37,28 +45,151 @@ def orientation(simplex):
     return sign
 
 
-def uniform_loss(simplex, ys=None):
+def uniform_loss(simplex, values, value_scale):
+    """
+    Uniform loss.
+
+    Parameters
+    ----------
+    simplex : list of tuples
+        Each entry is one point of the simplex.
+    values : list of values
+        The scaled function values of each of the simplex points.
+    value_scale : float
+        The scale of values, where ``values = function_values * value_scale``.
+
+    Returns
+    -------
+    loss : float
+    """
     return volume(simplex)
 
 
-def std_loss(simplex, ys):
-    r = np.linalg.norm(np.std(ys, axis=0))
+def std_loss(simplex, values, value_scale):
+    """
+    Computes the loss of the simplex based on the standard deviation.
+
+    Parameters
+    ----------
+    simplex : list of tuples
+        Each entry is one point of the simplex.
+    values : list of values
+        The scaled function values of each of the simplex points.
+    value_scale : float
+        The scale of values, where ``values = function_values * value_scale``.
+
+    Returns
+    -------
+    loss : float
+    """
+
+    r = np.linalg.norm(np.std(values, axis=0))
     vol = volume(simplex)
 
     dim = len(simplex) - 1
 
-    return r.flat * np.power(vol, 1. / dim) + vol
+    return r.flat * np.power(vol, 1.0 / dim) + vol
 
 
-def default_loss(simplex, ys):
-    # return std_loss(simplex, ys)
-    if isinstance(ys[0], Iterable):
-        pts = [(*x, *y) for x, y in zip(simplex, ys)]
+def default_loss(simplex, values, value_scale):
+    """
+    Computes the average of the volumes of the simplex.
+
+    Parameters
+    ----------
+    simplex : list of tuples
+        Each entry is one point of the simplex.
+    values : list of values
+        The scaled function values of each of the simplex points.
+    value_scale : float
+        The scale of values, where ``values = function_values * value_scale``.
+
+    Returns
+    -------
+    loss : float
+    """
+    if isinstance(values[0], Iterable):
+        pts = [(*x, *y) for x, y in zip(simplex, values)]
     else:
-        pts = [(*x, y) for x, y in zip(simplex, ys)]
+        pts = [(*x, y) for x, y in zip(simplex, values)]
     return simplex_volume_in_embedding(pts)
 
 
+@uses_nth_neighbors(1)
+def triangle_loss(simplex, values, value_scale, neighbors, neighbor_values):
+    """
+    Computes the average of the volumes of the simplex combined with each
+    neighbouring point.
+
+    Parameters
+    ----------
+    simplex : list of tuples
+        Each entry is one point of the simplex.
+    values : list of values
+        The scaled function values of each of the simplex points.
+    value_scale : float
+        The scale of values, where ``values = function_values * value_scale``.
+    neighbors : list of tuples
+        The neighboring points of the simplex, ordered such that simplex[0]
+        exactly opposes neighbors[0], etc.
+    neighbor_values : list of values
+        The function values for each of the neighboring points.
+
+    Returns
+    -------
+    loss : float
+    """
+
+    neighbors = [n for n in neighbors if n is not None]
+    neighbor_values = [v for v in neighbor_values if v is not None]
+    if len(neighbors) == 0:
+        return 0
+
+    s = [(*x, *to_list(y)) for x, y in zip(simplex, values)]
+    n = [(*x, *to_list(y)) for x, y in zip(neighbors, neighbor_values)]
+
+    return sum(simplex_volume_in_embedding([*s, neighbor]) for neighbor in n) / len(
+        neighbors
+    )
+
+
+def curvature_loss_function(exploration=0.05):
+    # XXX: add doc-string!
+    @uses_nth_neighbors(1)
+    def curvature_loss(simplex, values, value_scale, neighbors, neighbor_values):
+        """Compute the curvature loss of a simplex.
+
+        Parameters
+        ----------
+        simplex : list of tuples
+            Each entry is one point of the simplex.
+        values : list of values
+            The scaled function values of each of the simplex points.
+        value_scale : float
+            The scale of values, where ``values = function_values * value_scale``.
+        neighbors : list of tuples
+            The neighboring points of the simplex, ordered such that simplex[0]
+            exactly opposes neighbors[0], etc.
+        neighbor_values : list of values
+            The scaled function values for each of the neighboring points.
+
+        Returns
+        -------
+        loss : float
+        """
+        dim = len(simplex[0])  # the number of coordinates
+        loss_input_volume = volume(simplex)
+
+        loss_curvature = triangle_loss(
+            simplex, values, value_scale, neighbors, neighbor_values
+        )
+        return (
+            loss_curvature + exploration * loss_input_volume ** ((2 + dim) / dim)
+        ) ** (1 / (2 + dim))
+
+    return curvature_loss
+
+
 def choose_point_in_simplex(simplex, transform=None):
     """Choose a new point in inside a simplex.
 
@@ -69,9 +200,10 @@ def choose_point_in_simplex(simplex, transform=None):
     Parameters
     ----------
     simplex : numpy array
-        The coordinates of a triangle with shape (N+1, N)
+        The coordinates of a triangle with shape (N+1, N).
     transform : N*N matrix
-        The multiplication to apply to the simplex before choosing the new point
+        The multiplication to apply to the simplex before choosing
+        the new point.
 
     Returns
     -------
@@ -90,8 +222,7 @@ def choose_point_in_simplex(simplex, transform=None):
     else:
         distances = scipy.spatial.distance.pdist(simplex)
         distance_matrix = scipy.spatial.distance.squareform(distances)
-        i, j = np.unravel_index(np.argmax(distance_matrix),
-                                distance_matrix.shape)
+        i, j = np.unravel_index(np.argmax(distance_matrix), distance_matrix.shape)
         point = (simplex[i, :] + simplex[j, :]) / 2
 
     if transform is not None:
@@ -163,9 +294,23 @@ class LearnerND(BaseLearner):
     def __init__(self, func, bounds, loss_per_simplex=None):
         self._vdim = None
         self.loss_per_simplex = loss_per_simplex or default_loss
+
+        if hasattr(self.loss_per_simplex, "nth_neighbors"):
+            if self.loss_per_simplex.nth_neighbors > 1:
+                raise NotImplementedError(
+                    "The provided loss function wants "
+                    "next-nearest neighboring simplices for the loss computation, "
+                    "this feature is not yet implemented, either use "
+                    "nth_neightbors = 0 or 1"
+                )
+            self.nth_neighbors = self.loss_per_simplex.nth_neighbors
+        else:
+            self.nth_neighbors = 0
+
         self.data = OrderedDict()
         self.pending_points = set()
 
+        self.bounds = bounds
         if isinstance(bounds, scipy.spatial.ConvexHull):
             hull_points = bounds.points[bounds.vertices]
             self._bounds_points = sorted(list(map(tuple, hull_points)))
@@ -192,7 +337,9 @@ class LearnerND(BaseLearner):
         # for the output
         self._min_value = None
         self._max_value = None
-        self._output_multiplier = 1 # If we do not know anything, do not scale the values
+        self._output_multiplier = (
+            1  # If we do not know anything, do not scale the values
+        )
         self._recompute_losses_factor = 1.1
 
         # create a private random number generator with fixed seed
@@ -236,10 +383,10 @@ class LearnerND(BaseLearner):
     def bounds_are_done(self):
         return all(p in self.data for p in self._bounds_points)
 
-    def ip(self):
+    def _ip(self):
         """A `scipy.interpolate.LinearNDInterpolator` instance
         containing the learner's data."""
-        # XXX: take our own triangulation into account when generating the ip
+        # XXX: take our own triangulation into account when generating the _ip
         return interpolate.LinearNDInterpolator(self.points, self.values)
 
     @property
@@ -251,14 +398,15 @@ class LearnerND(BaseLearner):
 
         try:
             self._tri = Triangulation(self.points)
-            self.update_losses(set(), self._tri.simplices)
-            return self._tri
         except ValueError:
             # A ValueError is raised if we do not have enough points or
             # the provided points are coplanar, so we need more points to
             # create a valid triangulation
             return None
 
+        self._update_losses(set(), self._tri.simplices)
+        return self._tri
+
     @property
     def values(self):
         """Get the values from `data` as a numpy array."""
@@ -290,9 +438,8 @@ class LearnerND(BaseLearner):
             simplex = self._pending_to_simplex.get(point)
             if simplex is not None and not self._simplex_exists(simplex):
                 simplex = None
-            to_delete, to_add = tri.add_point(
-                point, simplex, transform=self._transform)
-            self.update_losses(to_delete, to_add)
+            to_delete, to_add = tri.add_point(point, simplex, transform=self._transform)
+            self._update_losses(to_delete, to_add)
 
     def _simplex_exists(self, simplex):
         simplex = tuple(sorted(simplex))
@@ -300,12 +447,13 @@ class LearnerND(BaseLearner):
 
     def inside_bounds(self, point):
         """Check whether a point is inside the bounds."""
-        if hasattr(self, '_interior'):
+        if hasattr(self, "_interior"):
             return self._interior.find_simplex(point, tol=1e-8) >= 0
         else:
             eps = 1e-8
-            return all((mn - eps) <= p <= (mx + eps) for p, (mn, mx)
-                       in zip(point, self._bbox))
+            return all(
+                (mn - eps) <= p <= (mx + eps) for p, (mn, mx) in zip(point, self._bbox)
+            )
 
     def tell_pending(self, point, *, simplex=None):
         point = tuple(point)
@@ -325,10 +473,10 @@ class LearnerND(BaseLearner):
 
         simplex = tuple(simplex)
         simplices = [self.tri.vertex_to_simplices[i] for i in simplex]
-        neighbours = set.union(*simplices)
+        neighbors = set.union(*simplices)
         # Neighbours also includes the simplex itself
 
-        for simpl in neighbours:
+        for simpl in neighbors:
             _, to_add = self._try_adding_pending_point_to_simplex(point, simpl)
             if to_add is None:
                 continue
@@ -369,8 +517,11 @@ class LearnerND(BaseLearner):
 
     def _ask_bound_point(self):
         # get the next bound point that is still available
-        new_point = next(p for p in self._bounds_points
-                         if p not in self.data and p not in self.pending_points)
+        new_point = next(
+            p
+            for p in self._bounds_points
+            if p not in self.data and p not in self.pending_points
+        )
         self.tell_pending(new_point)
         return new_point, np.inf
 
@@ -393,14 +544,19 @@ class LearnerND(BaseLearner):
         # find the simplex with the highest loss, we do need to check that the
         # simplex hasn't been deleted yet
         while len(self._simplex_queue):
+            # XXX: Need to add check that the loss is the most recent computed loss
             loss, simplex, subsimplex = self._simplex_queue.pop(0)
-            if (subsimplex is None
+            if (
+                subsimplex is None
                 and simplex in self.tri.simplices
-                and simplex not in self._subtriangulations):
+                and simplex not in self._subtriangulations
+            ):
                 return abs(loss), simplex, subsimplex
-            if (simplex in self._subtriangulations
+            if (
+                simplex in self._subtriangulations
                 and simplex in self.tri.simplices
-                and subsimplex in self._subtriangulations[simplex].simplices):
+                and subsimplex in self._subtriangulations[simplex].simplices
+            ):
                 return abs(loss), simplex, subsimplex
 
         # Could not find a simplex, this code should never be reached
@@ -423,8 +579,7 @@ class LearnerND(BaseLearner):
             subtri = self._subtriangulations[simplex]
             points = subtri.get_vertices(subsimplex)
 
-        point_new = tuple(choose_point_in_simplex(points,
-                                                  transform=self._transform))
+        point_new = tuple(choose_point_in_simplex(points, transform=self._transform))
 
         self._pending_to_simplex[point_new] = simplex
         self.tell_pending(point_new, simplex=simplex)  # O(??)
@@ -433,8 +588,10 @@ class LearnerND(BaseLearner):
 
     @property
     def _bounds_available(self):
-        return any((p not in self.pending_points and p not in self.data)
-                   for p in self._bounds_points)
+        return any(
+            (p not in self.pending_points and p not in self.data)
+            for p in self._bounds_points
+        )
 
     def _ask(self):
         if self._bounds_available:
@@ -448,7 +605,46 @@ class LearnerND(BaseLearner):
 
         return self._ask_best_point()  # O(log N)
 
-    def update_losses(self, to_delete: set, to_add: set):
+    def _compute_loss(self, simplex):
+        # get the loss
+        vertices = self.tri.get_vertices(simplex)
+        values = [self.data[tuple(v)] for v in vertices]
+
+        # scale them to a cube with sides 1
+        vertices = vertices @ self._transform
+        values = self._output_multiplier * np.array(values)
+
+        if self.nth_neighbors == 0:
+            # compute the loss on the scaled simplex
+            return float(
+                self.loss_per_simplex(vertices, values, self._output_multiplier)
+            )
+
+        # We do need the neighbors
+        neighbors = self.tri.get_opposing_vertices(simplex)
+
+        neighbor_points = self.tri.get_vertices(neighbors)
+        neighbor_values = [self.data.get(x, None) for x in neighbor_points]
+
+        for i, point in enumerate(neighbor_points):
+            if point is not None:
+                neighbor_points[i] = point @ self._transform
+
+        for i, value in enumerate(neighbor_values):
+            if value is not None:
+                neighbor_values[i] = self._output_multiplier * value
+
+        return float(
+            self.loss_per_simplex(
+                vertices,
+                values,
+                self._output_multiplier,
+                neighbor_points,
+                neighbor_values,
+            )
+        )
+
+    def _update_losses(self, to_delete: set, to_add: set):
         # XXX: add the points outside the triangulation to this as well
         pending_points_unbound = set()
 
@@ -458,11 +654,11 @@ class LearnerND(BaseLearner):
             if subtri is not None:
                 pending_points_unbound.update(subtri.vertices)
 
-        pending_points_unbound = set(p for p in pending_points_unbound
-                                     if p not in self.data)
-
+        pending_points_unbound = {
+            p for p in pending_points_unbound if p not in self.data
+        }
         for simplex in to_add:
-            loss = self.compute_loss(simplex)
+            loss = self._compute_loss(simplex)
             self._losses[simplex] = loss
 
             for p in pending_points_unbound:
@@ -473,21 +669,28 @@ class LearnerND(BaseLearner):
                 continue
 
             self._update_subsimplex_losses(
-                simplex, self._subtriangulations[simplex].simplices)
+                simplex, self._subtriangulations[simplex].simplices
+            )
 
-    def compute_loss(self, simplex):
-        # get the loss
-        vertices = self.tri.get_vertices(simplex)
-        values = [self.data[tuple(v)] for v in vertices]
+        if self.nth_neighbors:
+            points_of_added_simplices = set.union(*[set(s) for s in to_add])
+            neighbors = (
+                self.tri.get_simplices_attached_to_points(points_of_added_simplices)
+                - to_add
+            )
+            for simplex in neighbors:
+                loss = self._compute_loss(simplex)
+                self._losses[simplex] = loss
 
-        # scale them to a cube with sides 1
-        vertices = vertices @ self._transform
-        values = self._output_multiplier * values
+                if simplex not in self._subtriangulations:
+                    self._simplex_queue.add((loss, simplex, None))
+                    continue
 
-        # compute the loss on the scaled simplex
-        return float(self.loss_per_simplex(vertices, values))
+                self._update_subsimplex_losses(
+                    simplex, self._subtriangulations[simplex].simplices
+                )
 
-    def recompute_all_losses(self):
+    def _recompute_all_losses(self):
         """Recompute all losses and pending losses."""
         # amortized O(N) complexity
         if self.tri is None:
@@ -498,7 +701,7 @@ class LearnerND(BaseLearner):
 
         # recompute all losses
         for simplex in self.tri.simplices:
-            loss = self.compute_loss(simplex)
+            loss = self._compute_loss(simplex)
             self._losses[simplex] = loss
 
             # now distribute it around the the children if they are present
@@ -507,7 +710,8 @@ class LearnerND(BaseLearner):
                 continue
 
             self._update_subsimplex_losses(
-                simplex, self._subtriangulations[simplex].simplices)
+                simplex, self._subtriangulations[simplex].simplices
+            )
 
     @property
     def _scale(self):
@@ -517,59 +721,43 @@ class LearnerND(BaseLearner):
     def _update_range(self, new_output):
         if self._min_value is None or self._max_value is None:
             # this is the first point, nothing to do, just set the range
-            self._min_value = np.array(new_output)
-            self._max_value = np.array(new_output)
-            self._old_scale = self._scale
+            self._min_value = np.min(new_output)
+            self._max_value = np.max(new_output)
+            self._old_scale = self._scale or 1
             return False
 
         # if range in one or more directions is doubled, then update all losses
-        self._min_value = np.minimum(self._min_value, new_output)
-        self._max_value = np.maximum(self._max_value, new_output)
+        self._min_value = min(self._min_value, np.min(new_output))
+        self._max_value = max(self._max_value, np.max(new_output))
 
-        scale_multiplier = 1 / self._scale
-        if isinstance(scale_multiplier, float):
-            scale_multiplier = np.array([scale_multiplier], dtype=float)
+        scale_multiplier = 1 / (self._scale or 1)
 
         # the maximum absolute value that is in the range. Because this is the
         # largest number, this also has the largest absolute numerical error.
-        max_absolute_value_in_range = np.max(np.abs([self._min_value, self._max_value]), axis=0)
+        max_absolute_value_in_range = max(abs(self._min_value), abs(self._max_value))
         # since a float has a relative error of 1e-15, the absolute error is the value * 1e-15
         abs_err = 1e-15 * max_absolute_value_in_range
         # when scaling the floats, the error gets increased.
         scaled_err = abs_err * scale_multiplier
 
-        allowed_numerical_error = 1e-2
-
         # do not scale along the axis if the numerical error gets too big
-        scale_multiplier[scaled_err > allowed_numerical_error] = 1
+        if scaled_err > 1e-2:  # allowed_numerical_error = 1e-2
+            scale_multiplier = 1
 
         self._output_multiplier = scale_multiplier
 
-        scale_factor = np.max(np.nan_to_num(self._scale / self._old_scale))
+        scale_factor = self._scale / self._old_scale
         if scale_factor > self._recompute_losses_factor:
             self._old_scale = self._scale
-            self.recompute_all_losses()
+            self._recompute_all_losses()
             return True
         return False
 
-    def losses(self):
-        """Get the losses of each simplex in the current triangulation, as dict
-
-        Returns
-        -------
-        losses : dict
-            the key is a simplex, the value is the loss of this simplex
-        """
-        # XXX could be a property
-        if self.tri is None:
-            return dict()
-
-        return self._losses
-
     @cache_latest
     def loss(self, real=True):
-        losses = self.losses()  # XXX: compute pending loss if real == False
-        return max(losses.values()) if losses else float('inf')
+        # XXX: compute pending loss if real == False
+        losses = self._losses if self.tri is not None else dict()
+        return max(losses.values()) if losses else float("inf")
 
     def remove_unfinished(self):
         # XXX: implement this method
@@ -593,11 +781,14 @@ class LearnerND(BaseLearner):
         """
         hv = ensure_holoviews()
         if self.vdim > 1:
-            raise NotImplementedError('holoviews currently does not support',
-                                      '3D surface plots in bokeh.')
+            raise NotImplementedError(
+                "holoviews currently does not support", "3D surface plots in bokeh."
+            )
         if self.ndim != 2:
-            raise NotImplementedError("Only 2D plots are implemented: You can "
-                                      "plot a 2D slice with 'plot_slice'.")
+            raise NotImplementedError(
+                "Only 2D plots are implemented: You can "
+                "plot a 2D slice with 'plot_slice'."
+            )
         x, y = self._bbox
         lbrt = x[0], y[0], x[1], y[1]
 
@@ -612,17 +803,20 @@ class LearnerND(BaseLearner):
             xs = ys = np.linspace(0, 1, n)
             xs = xs * (x[1] - x[0]) + x[0]
             ys = ys * (y[1] - y[0]) + y[0]
-            z = self.ip()(xs[:, None], ys[None, :]).squeeze()
+            z = self._ip()(xs[:, None], ys[None, :]).squeeze()
 
             im = hv.Image(np.rot90(z), bounds=lbrt)
 
             if tri_alpha:
-                points = np.array([self.tri.get_vertices(s)
-                                   for s in self.tri.simplices])
-                points = np.pad(points[:, [0, 1, 2, 0], :],
-                                pad_width=((0, 0), (0, 1), (0, 0)),
-                                mode='constant',
-                                constant_values=np.nan).reshape(-1, 2)
+                points = np.array(
+                    [self.tri.get_vertices(s) for s in self.tri.simplices]
+                )
+                points = np.pad(
+                    points[:, [0, 1, 2, 0], :],
+                    pad_width=((0, 0), (0, 1), (0, 0)),
+                    mode="constant",
+                    constant_values=np.nan,
+                ).reshape(-1, 2)
                 tris = hv.EdgePaths([points])
             else:
                 tris = hv.EdgePaths([])
@@ -630,7 +824,7 @@ class LearnerND(BaseLearner):
             im = hv.Image([], bounds=lbrt)
             tris = hv.EdgePaths([])
 
-        im_opts = dict(cmap='viridis')
+        im_opts = dict(cmap="viridis")
         tri_opts = dict(line_width=0.5, alpha=tri_alpha)
         no_hover = dict(plot=dict(inspection_policy=None, tools=[]))
 
@@ -654,14 +848,17 @@ class LearnerND(BaseLearner):
             if not self.data:
                 return hv.Scatter([]) * hv.Path([])
             elif self.vdim > 1:
-                raise NotImplementedError('multidimensional output not yet'
-                                          ' supported by `plot_slice`')
+                raise NotImplementedError(
+                    "multidimensional output not yet supported by `plot_slice`"
+                )
             n = n or 201
-            values = [cut_mapping.get(i, np.linspace(*self._bbox[i], n))
-                      for i in range(self.ndim)]
+            values = [
+                cut_mapping.get(i, np.linspace(*self._bbox[i], n))
+                for i in range(self.ndim)
+            ]
             ind = next(i for i in range(self.ndim) if i not in cut_mapping)
             x = values[ind]
-            y = self.ip()(*values)
+            y = self._ip()(*values)
             p = hv.Path((x, y))
 
             # Plot with 5% margins such that the boundary points are visible
@@ -671,8 +868,9 @@ class LearnerND(BaseLearner):
 
         elif plot_dim == 2:
             if self.vdim > 1:
-                raise NotImplementedError('holoviews currently does not support'
-                                          ' 3D surface plots in bokeh.')
+                raise NotImplementedError(
+                    "holoviews currently does not support 3D surface plots in bokeh."
+                )
             if n is None:
                 # Calculate how many grid points are needed.
                 # factor from A=√3/4 * a² (equilateral triangle)
@@ -682,21 +880,23 @@ class LearnerND(BaseLearner):
 
             xs = ys = np.linspace(0, 1, n)
             xys = [xs[:, None], ys[None, :]]
-            values = [cut_mapping[i] if i in cut_mapping
-                      else xys.pop(0) * (b[1] - b[0]) + b[0]
-                      for i, b in enumerate(self._bbox)]
+            values = [
+                cut_mapping[i]
+                if i in cut_mapping
+                else xys.pop(0) * (b[1] - b[0]) + b[0]
+                for i, b in enumerate(self._bbox)
+            ]
 
-            lbrt = [b for i, b in enumerate(self._bbox)
-                    if i not in cut_mapping]
+            lbrt = [b for i, b in enumerate(self._bbox) if i not in cut_mapping]
             lbrt = np.reshape(lbrt, (2, 2)).T.flatten().tolist()
 
             if len(self.data) >= 4:
-                z = self.ip()(*values).squeeze()
+                z = self._ip()(*values).squeeze()
                 im = hv.Image(np.rot90(z), bounds=lbrt)
             else:
                 im = hv.Image([], bounds=lbrt)
 
-            return im.opts(style=dict(cmap='viridis'))
+            return im.opts(style=dict(cmap="viridis"))
         else:
             raise ValueError("Only 1 or 2-dimensional plots can be generated.")
 
@@ -729,23 +929,38 @@ class LearnerND(BaseLearner):
                     Ye += [vertices[i][1] for i in s] + [None]
                     Ze += [vertices[i][2] for i in s] + [None]
 
-            plots.append(plotly.graph_objs.Scatter3d(
-                x=Xe, y=Ye, z=Ze, mode='lines',
-                line=dict(color='rgb(125,125,125)', width=1),
-                hoverinfo='none'
-            ))
+            plots.append(
+                plotly.graph_objs.Scatter3d(
+                    x=Xe,
+                    y=Ye,
+                    z=Ze,
+                    mode="lines",
+                    line=dict(color="rgb(125,125,125)", width=1),
+                    hoverinfo="none",
+                )
+            )
 
         Xn, Yn, Zn = zip(*vertices)
         colors = [self.data[p] for p in self.tri.vertices]
-        marker = dict(symbol='circle', size=3, color=colors,
-            colorscale='Viridis',
-            line=dict(color='rgb(50,50,50)', width=0.5))
+        marker = dict(
+            symbol="circle",
+            size=3,
+            color=colors,
+            colorscale="Viridis",
+            line=dict(color="rgb(50,50,50)", width=0.5),
+        )
 
-        plots.append(plotly.graph_objs.Scatter3d(
-            x=Xn, y=Yn, z=Zn, mode='markers',
-            name='actors', marker=marker,
-            hoverinfo='text'
-        ))
+        plots.append(
+            plotly.graph_objs.Scatter3d(
+                x=Xn,
+                y=Yn,
+                z=Zn,
+                mode="markers",
+                name="actors",
+                marker=marker,
+                hoverinfo="text",
+            )
+        )
 
         axis = dict(
             showbackground=False,
@@ -753,14 +968,15 @@ class LearnerND(BaseLearner):
             zeroline=False,
             showgrid=False,
             showticklabels=False,
-            title='',
+            title="",
         )
 
         layout = plotly.graph_objs.Layout(
             showlegend=False,
             scene=dict(xaxis=axis, yaxis=axis, zaxis=axis),
             margin=dict(t=100),
-            hovermode='closest')
+            hovermode="closest",
+        )
 
         fig = plotly.graph_objs.Figure(data=plots, layout=layout)
 
@@ -770,19 +986,23 @@ class LearnerND(BaseLearner):
         return self.data
 
     def _set_data(self, data):
-        self.tell_many(*zip(*data.items()))
+        if data:
+            self.tell_many(*zip(*data.items()))
 
-    def _get_iso(self, level=0.0, which='surface'):
-        if which == 'surface':
+    def _get_iso(self, level=0.0, which="surface"):
+        if which == "surface":
             if self.ndim != 3 or self.vdim != 1:
-                raise Exception('Isosurface plotting is only supported'
-                                ' for a 3D input and 1D output')
+                raise Exception(
+                    "Isosurface plotting is only supported"
+                    " for a 3D input and 1D output"
+                )
             get_surface = True
             get_line = False
-        elif which == 'line':
+        elif which == "line":
             if self.ndim != 2 or self.vdim != 1:
-                raise Exception('Isoline plotting is only supported'
-                                ' for a 2D input and 1D output')
+                raise Exception(
+                    "Isoline plotting is only supported for a 2D input and 1D output"
+                )
             get_surface = False
             get_line = True
 
@@ -800,8 +1020,7 @@ class LearnerND(BaseLearner):
             db = abs(value_b - level)
             dab = da + db
 
-            new_pt = (db / dab * np.array(vertex_a)
-                      + da / dab * np.array(vertex_b))
+            new_pt = db / dab * np.array(vertex_a) + da / dab * np.array(vertex_b)
 
             new_index = len(vertices)
             vertices.append(new_pt)
@@ -873,11 +1092,11 @@ class LearnerND(BaseLearner):
                 plot = plot * self.plot_isoline(level=l, n=-1)
             return plot
 
-        vertices, lines = self._get_iso(level, which='line')
+        vertices, lines = self._get_iso(level, which="line")
         paths = [[vertices[i], vertices[j]] for i, j in lines]
         contour = hv.Path(paths)
 
-        contour_opts = dict(color='black')
+        contour_opts = dict(color="black")
         contour = contour.opts(style=contour_opts)
         return plot * contour
 
@@ -901,15 +1120,16 @@ class LearnerND(BaseLearner):
         """
         plotly = ensure_plotly()
 
-        vertices, faces = self._get_iso(level, which='surface')
+        vertices, faces = self._get_iso(level, which="surface")
         x, y, z = zip(*vertices)
 
         fig = plotly.figure_factory.create_trisurf(
-            x=x, y=y, z=z, plot_edges=False,
-            simplices=faces, title="Isosurface")
+            x=x, y=y, z=z, plot_edges=False, simplices=faces, title="Isosurface"
+        )
         isosurface = fig.data[0]
-        isosurface.update(lighting=dict(ambient=1, diffuse=1,
-            roughness=1, specular=0, fresnel=0))
+        isosurface.update(
+            lighting=dict(ambient=1, diffuse=1, roughness=1, specular=0, fresnel=0)
+        )
 
         if hull_opacity < 1e-3:
             # Do not compute the hull_mesh.
@@ -937,16 +1157,22 @@ class LearnerND(BaseLearner):
                     return color
                 if scipy.spatial.ConvexHull(points).volume < 1e-5:
                     return color
-            color_dict[simplex] = tuple(random.randint(0, 255)
-                                        for _ in range(3))
+            color_dict[simplex] = tuple(random.randint(0, 255) for _ in range(3))
             return color_dict[simplex]
 
         colors = [_get_plane_color(simplex) for simplex in hull.simplices]
 
         x, y, z = zip(*self._bounds_points)
         i, j, k = hull.simplices.T
-        lighting = dict(ambient=1, diffuse=1, roughness=1,
-                        specular=0, fresnel=0)
-        return plotly.graph_objs.Mesh3d(x=x, y=y, z=z, i=i, j=j, k=k,
-                                        facecolor=colors, opacity=opacity,
-                                        lighting=lighting)
+        lighting = dict(ambient=1, diffuse=1, roughness=1, specular=0, fresnel=0)
+        return plotly.graph_objs.Mesh3d(
+            x=x,
+            y=y,
+            z=z,
+            i=i,
+            j=j,
+            k=k,
+            facecolor=colors,
+            opacity=opacity,
+            lighting=lighting,
+        )
